/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _COGLIB_H_INCLUDED_
#define _COGLIB_H_INCLUDED_

#define COGLIBAPI static Sith::COG::VMType* __cdecl
#define COGLIB Sith::COG::VMType* __cdecl

namespace Sith
{
	namespace COG
	{
		// CogLib Class
		// - Defines and implements the COGLIB/COG Library
		// - Tracks, stores, and calls COG verbs
		class CogLib
		{
		private:
			class LibElement
			{
			public:
				std::string Name;
				int NumArgs;

				void *libftn;

				LibElement(std::string _Name, int _NumArgs, void* _libftn) : Name(_Name), NumArgs(_NumArgs), libftn(_libftn)
				{
					return;
				};
			};

			static std::vector<LibElement> Library;
			static int LibraryCount;

			static UniTable COGStrings;
			static UniTable CreditStrings;
			static UniTable CutStrings;
			static UniTable JKStrings;
			static UniTable SithStrings;

		public:
			static inline SITHERROR LoadCOGStrings(std::string FileName)
			{
				COGStrings.Clear();
				return COGStrings.Load(FileName);
			};

			static inline SITHERROR LoadCreditStrings(std::string FileName)
			{
				CreditStrings.Clear();
				return CreditStrings.Load(FileName);
			};

			static inline SITHERROR LoadCutStrings(std::string FileName)
			{
				CutStrings.Clear();
				return CutStrings.Load(FileName);
			};

			static inline SITHERROR LoadJKStrings(std::string FileName)
			{
				JKStrings.Clear();
				return JKStrings.Load(FileName);
			};

			static inline SITHERROR LoadSithStrings(std::string FileName)
			{
				SithStrings.Clear();
				return SithStrings.Load(FileName);
			};

			static inline std::string GetUniString(std::string Key)
			{
				std::string RetVal;
				RetVal = COGStrings.Get(Key);
				if(RetVal.length() > 0) return RetVal;

				RetVal = JKStrings.Get(Key);
				if(RetVal.length() > 0) return RetVal;

				RetVal = SithStrings.Get(Key);
				if(RetVal.length() > 0) return RetVal;

				return Key;
			};

			// TranslateLibraryName()
			// - Takes a string input and an argument count. Returns the correct libcall argument if it's a correct function
			static inline int TranslateLibraryName(std::string _Name, int _NumArgs)
			{
				std::transform(_Name.begin(), _Name.end(), _Name.begin(), tolower);
				int LibcallOffset = 0;
				for(std::vector<LibElement>::iterator it = Library.begin(); it != Library.end(); ++it)
				{
					if((it->Name == _Name) && (it->NumArgs == _NumArgs))
					{
						return LibcallOffset;
					};

					++LibcallOffset;
				};

				return -1;
			};

			// InjectLibraryFunction()
			// - Adds a new library function to the COGLIB
			static inline void InjectLibraryFunction(std::string _Name, int _NumArgs, void* _libftn)
			{
				std::transform(_Name.begin(), _Name.end(), _Name.begin(), tolower);
				for(std::vector<LibElement>::iterator it = Library.begin(); it != Library.end(); ++it)
				{
					if((it->Name == _Name) && (it->NumArgs == _NumArgs))
					{
						return;
					};
				};

				Library.push_back(LibElement(_Name, _NumArgs, _libftn));
				++LibraryCount;
				return;
			};

			// Call()
			// - Calls a library function by LIBCALL offset
			static inline void Call(int Offset, StackMachine* GSM)
			{
				if(Offset < 0 || Offset >= LibraryCount)
				{
					return;
				}
				else
				{
					int NumArgs = Library[Offset].NumArgs;
					size_t VMTSize = sizeof(VMType);

					VMType *Arg = new VMType[NumArgs];
					for(int i = 0; i < NumArgs; ++i)
					{
						Arg[i] = VMType(GSM->Stack.top());
						GSM->Stack.pop();
					};

					VMType* ReturnValue;

					void* ftn = Library[Offset].libftn;

					// ********************** CAUTION: x86-specific!!!! *************************
					// * This is a custom implementation of CDECL.								*
					// * That means this will not work on platforms which depend on different	*
					// * calling conventions. Under Windows, __cdecl must be explicitly defined.*
					// * This should be handled by the COGLIBAPI macro. If you don't, it will	*
					// * use "stdcall" *snicker* by default.									*
					// **************************************************************************
					__asm
					{
						push GSM
						mov eax, NumArgs
						mul VMTSize
						mov ecx, eax
						xor ebx, ebx
					StartLoop:
						cmp ecx, 0
						je EndStart
						mov eax, Arg
						add eax, ebx
						push eax
						sub ecx, VMTSize
						add ebx, VMTSize
						jmp StartLoop
					EndStart:

						call ftn

						mov [ReturnValue], eax

						mov ebx, NumArgs
						xor ecx, ecx
					EndLoop:
						cmp ecx, ebx
						je EndEnd
						pop eax
						add ecx, 1
						cmp ecx, ebx
						jne EndLoop
					EndEnd:
						pop eax
					};

					// **** END OF x86-specific ******

					GSM->ReturnRegister = VMType(*ReturnValue);

					delete ReturnValue;
					delete[] Arg;
				};

				return;
			};

			static inline SITHERROR InjectNamespace(std::string Name)
			{
				Sith::COG::VMHeap::AddNamespace(Name);
				return SE_OK;
			};

			static inline SITHERROR AddSetting(std::string Name, VMType Default)
			{
				std::string ScopeName;
				std::string VarName;
				unsigned int i = 0;

				for(i = 0; i < Name.length(); ++i)
				{
					if((Name[i] == ':') && (Name[i + 1] == ':'))
					{
						i += 2;
						break;
					}
					else
					{
						ScopeName.push_back(Name[i]);
					};
				};

				for(; i < Name.length(); ++i)
				{
					VarName.push_back(Name[i]);
				};

				int ScopeAddr = Sith::COG::VMHeap::TranslateNamespaceName(ScopeName);
				if(ScopeAddr == -1)
				{
					std::ostringstream Stream;

					Stream << "WARNING: Could not add setting \'" << Name << "\', namespace \'" << ScopeName << "\' does not exist.\n";
					ConOut(Stream.str(), ConOut::Warning);

					return SE_GENERAL_WARNING;
				};

				Sith::COG::VMHeap::AddGlobal(ScopeAddr, VarName, Default);

				return SE_OK;
			};

			static inline VMType GetSetting(std::string Name)
			{
				std::string ScopeName;
				std::string VarName;
				unsigned int i = 0;

				for(i = 0; i < Name.length(); ++i)
				{
					if((Name[i] == ':') && (Name[i + 1] == ':'))
					{
						i += 2;
						break;
					}
					else
					{
						ScopeName.push_back(Name[i]);
					};
				};

				for(; i < Name.length(); ++i)
				{
					VarName.push_back(Name[i]);
				};

				int ScopeAddr = Sith::COG::VMHeap::TranslateNamespaceName(ScopeName);
				if(ScopeAddr == -1)
				{
					std::ostringstream Stream;

					Stream << "WARNING: Could not retrieve setting \'" << Name << "\', namespace \'" << ScopeName << "\' does not exist.\n";
					ConOut(Stream.str(), ConOut::Warning);

					return VMType(VMTC_VOID);
				};

				int VarAddr = Sith::COG::VMHeap::TranslateNamespaceHeapName(ScopeAddr, VarName);
				if(VarAddr == -1)
				{
					std::ostringstream Stream;

					Stream << "WARNING: Could not retrieve setting \'" << Name << "\', symbol \'" << VarName << "\' is not a member of namespace \'" << ScopeName << "\'.\n";
					ConOut(Stream.str(), ConOut::Warning);

					return VMType(VMTC_VOID);
				};

				return Sith::COG::VMHeap::GetGlobal(VarAddr, ScopeAddr);
			};

			COGLIBAPI lscmd()
			{
				std::ostringstream Stream;
				Stream << "The COG Library currently registers " << (unsigned int)CogLib::Library.size() << " verbs.\n";
				ConOut(Stream.str(), ConOut::Standard);

				for(std::vector<LibElement>::iterator it = CogLib::Library.begin(); it != CogLib::Library.end(); ++it)
				{
					Stream.str("");
					Stream << "\t| " << it->Name << "(" << it->NumArgs << ");\n";
					ConOut(Stream.str(), ConOut::Standard);
				};

				ConOut("End of list\n\n", ConOut::Standard);

				return new COG::VMType(COG::VMTC_VOID);
			};

			// Vector arithmetic bindings:
			COGLIBAPI VectorAdd(COG::VMType* Vec1, COG::VMType* Vec2)
			{
				return new COG::VMType(Math::VectorAdd(Vec1->GetVector(), Vec2->GetVector()));
			};

			COGLIBAPI VectorCross(COG::VMType* Vec1, COG::VMType* Vec2)
			{
				return new COG::VMType(Math::VectorCross(Vec1->GetVector(), Vec2->GetVector()));
			};

			COGLIBAPI VectorDist(COG::VMType* Vec1, COG::VMType* Vec2)
			{
				return new COG::VMType(Math::VectorDist(Vec1->GetVector(), Vec2->GetVector()));
			};

			COGLIBAPI VectorDistSq(COG::VMType* Vec1, COG::VMType *Vec2)
			{
				return new COG::VMType(Math::VectorDistSq(Vec1->GetVector(), Vec2->GetVector()));
			};

			COGLIBAPI VectorDot(COG::VMType* Vec1, COG::VMType *Vec2)
			{
				return new COG::VMType(Math::VectorDot(Vec1->GetVector(), Vec2->GetVector()));
			};

			COGLIBAPI VectorLen(COG::VMType* Vec1)
			{
				return new COG::VMType(Math::VectorLen(Vec1->GetVector()));
			};

			COGLIBAPI VectorLenSq(COG::VMType* Vec1)
			{
				return new COG::VMType(Math::VectorLenSq(Vec1->GetVector()));
			};

			COGLIBAPI VectorNorm(COG::VMType* Vec1)
			{
				return new COG::VMType(Math::VectorNorm(Vec1->GetVector()));
			};

			COGLIBAPI VectorScale(COG::VMType* Vec1, COG::VMType* Float)
			{
				return new COG::VMType(Math::VectorScale(Vec1->GetVector(), Float->GetFloat()));
			};

			COGLIBAPI VectorSet(COG::VMType* X, COG::VMType* Y, COG::VMType* Z)
			{
				return new COG::VMType(Math::VectorSet(X->GetFloat(), Y->GetFloat(), Z->GetFloat()));
			};

			COGLIBAPI VectorSub(COG::VMType* Vec1, COG::VMType* Vec2)
			{
				return new COG::VMType(Math::VectorSub(Vec1->GetVector(), Vec2->GetVector()));
			};

			COGLIBAPI VectorX(COG::VMType* Vec1)
			{
				return new COG::VMType(Math::VectorX(Vec1->GetVector()));
			};

			COGLIBAPI VectorY(COG::VMType* Vec1)
			{
				return new COG::VMType(Math::VectorY(Vec1->GetVector()));
			};

			COGLIBAPI VectorZ(COG::VMType* Vec1)
			{
				return new COG::VMType(Math::VectorZ(Vec1->GetVector()));
			};
		};
	};
};
#endif // _COGLIB_H_INCLUDED_
