/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _OGL_RENDER_H_INCLUDED_
#define _OGL_RENDER_H_INCLUDED_

#include "ogl_hwbm.h"
#pragma comment (lib, "opengl32")
#include <gl.h>
#include <glext.h>
#include <wglext.h>
#include "ogl_extensions.h"
#include "ogl_render_compiled.h"

#include "worldstate.h"

#define PIOVER180 0.0174532925f

namespace Sith
{
	namespace Win32
	{
		class WaterSim
		{
		public:

			float *heightmap;
			float *normalmap;
			unsigned char *normaltexdata;

			unsigned int normaltex;

			float gravity;
			float wind[2];

			int heightdim;
			float heightextx;
			float heightexty;
			float waveamp;

			float normalmapscale;

			float elapsedtime;

			WaterSim()
			{
				heightmap = 0;
				normalmap = 0;
				normaltexdata = 0;
				h0r = 0;
				h0i = 0;
				kv = 0;
				skg = 0;
				workarea1 = 0;
				workarea2 = 0;
				workarea3 = 0;

				normaltex = 0;
			}

			~WaterSim()
			{
				Shutdown();
			}

			void Step(float timestep);
			void Start(int seed, int dim, float sizex, float sizey, float grav, float windx, float windy, float wavesize, float normalscale);
			void Shutdown();

			int GetDim()
			{
				return heightdim;
			}

		private:
			float *h0r;
			float *h0i;
			float *kv;
			float *skg;
			float *workarea1;
			int *workarea2;
			float *workarea3;

			float Phillips(float a, float k[2], float wind[2]);
		};

		struct OGLUIVertex
		{
			float p[3]; // position
			unsigned char c[4]; // color
			float t[2]; // texcoord
		};

		struct OGLCamera
		{
			float x, y, z, pitch, yaw;
			float forward[3], right[3], up[3];
			int sector;
			VFS::WisePointer<Res::Model> POVModel;
			Math::Vector MulTint, AddTint;

			OGLCamera()
			{
				sector = -1;
				POVModel = 0;
			}
		};

		class ReflectTexInfo
		{
		public:

			unsigned int texid;

			int sizex, sizey;

			float normal[4]; // 4th component is distance

			std::vector<std::vector<Res::Level::SURFACE>::iterator> surface;

			ReflectTexInfo()
			{
				texid = 0;
				sizex = 0;
				sizey = 0;
			}
			~ReflectTexInfo()
			{
				DeleteTex();
			}

			void GenTex(int width, int height)
			{
				if (texid)
					glDeleteTextures(1, &texid);
				texid = 0;

				sizex = width;
				sizey = height;

				glGenTextures(1, &texid);
				glBindTexture(GL_TEXTURE_2D, texid);
				glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, sizex, sizey, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			}
			void DeleteTex()
			{
				//if (texid)
				//	glDeleteTextures(1, &texid);
				//texid = 0;
			}

		private:
		};

		class Renderer : public System::Task
		{
			friend class Sith::System::KernelProc;
		private:
			WNDCLASSEX wc;
			DEVMODE dmOriginalScreenSettings;
			HDC hDC;
			HGLRC hRC;
			MSG Msg;

			static Renderer* StaticInterface;

			Renderer();
			~Renderer();

			static LRESULT CALLBACK RenderWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

			SITHERROR Startup();
			SITHERROR Process();

			SITHERROR DrawUIWidget(UI::BaseWidget& Widget, UI::BaseWidget& Parent);

			void DrawLevel();
			void Render3doNode(VFS::WisePointer<Sith::Res::Model> model, int geoset, int node, int sector);
			void Render3do(VFS::WisePointer<Sith::Res::Model> model, int sector);
			// if fancy water is on, don't call this with drawwater = true more than once per frame, as the reflections update each time
			void DrawJKL(bool DrawWater, bool DrawSky, int startsector, std::vector<unsigned char> sectorvis);
			void OnResize(int width, int height, int xoffset, int yoffset);
			void Shutdown();
			Sith::SITHERROR CreateOGLWindow();

			void DrawLevelBasic();
			void DrawJKLBasic(int startsector, std::vector<unsigned char> sectorvis);
	
			void AdvancedStartup();
			void DrawLevelAdvanced();
			void DrawJKLAdvanced(bool DrawReflect, bool DrawSky, int startsector, std::vector<unsigned char> sectorvis);
			void UpdateReflections(std::vector<std::vector<Res::Level::SURFACE>::iterator> &reflectionlist);

			void DrawJKLGeneric(int startsector, std::vector<unsigned char> sectorvis);
			void DrawLevel2();

			void CompileLevel();


			// shader stuff
			unsigned int watervp, waterfp, reflectvp, reflectfp;
			int LoadVertexProgram(const char *filename);
			int LoadFragmentProgram(const char *filename);
			void ShaderStartup();

			// window stuff
            int wsizex, wsizey, wbitdepth, wdepthbits, wstencilbits, fullscreen;
			float nearclip;

			bool ClearEveryFrame;

//			bool DeviceSupportsBMs;	// Does device support hardware acceleration and scaling for non-pow2 GUI BMs? If not, this causes horrible problems.
			// Virtually all modern consumer hardware supports conditional nonpow2s so I'm not too worried about it. Well... sorta worried. Thus the bool.

//			std::vector<VFS::WisePointer<Res::Material> > matlist;

			Win32::Timer timer;
			float frametime;

			OGLCamera Camera;
			float modelview[16];
			float projection[16];

			// fancy water stuff
			WaterSim water;
			unsigned int refractiontex;
			unsigned int framebuffer;
			unsigned int renderbufferdepth;
			int framebuffersize;
			float waterfovmod;
			// for water sorting, planes below this distance from each other considered the same
			float reflection_tolerance;
			int reflection_buffer_size; // number of reflection textures to keep around
			std::vector<ReflectTexInfo> reflect_tex;
			// debug stuff
			int reflectionplanes;
			int sectors_drawn;
			int titleupdate;

			CompiledLevel clevel;

		public:
			HWND hMainWnd;

			// level pointer
			VFS::WisePointer<Res::Level> level;

			// anisotropy setting for new textures
			float anisotropy;

			int resetting; // to avoid closing sith2 during video reset
			void ReloadTextures();


			SITHERROR DrawUIText (std::string Text, unsigned int x, unsigned int y, VFS::WisePointer<Res::Font> Font, float Scale=1.0f);


			static inline Renderer& Get()
			{
				return *StaticInterface;
			};

			inline bool IsBMSupported()
			{
//				return DeviceSupportsBMs;
				return true;
			};

			void SetCamera(float x, float y, float z, float pitch, float yaw, int sector, Math::Vector MulTint, Math::Vector AddTint, VFS::WisePointer<Res::Model> _POVModel = 0)
			{
                Camera.POVModel = _POVModel;
				Camera.sector = sector;
				Camera.MulTint = MulTint;
				Camera.AddTint = AddTint;

				Camera.x = x;
				Camera.y = y;
				Camera.z = z;
				Camera.pitch = pitch;
				Camera.yaw = yaw;

				float xzscale = cos(Camera.pitch * PIOVER180);
				Camera.forward[0] = -sin(Camera.yaw * PIOVER180) * xzscale;
				Camera.forward[1] = cos(Camera.yaw * PIOVER180) * xzscale;
				Camera.forward[2] = sin(Camera.pitch * PIOVER180);

				Camera.right[0] = -sinf((Camera.yaw - 90) * PIOVER180);
				Camera.right[1] = cosf((Camera.yaw - 90) * PIOVER180);
				Camera.right[2] = 0;

				// up = right x forward
				Camera.up[0] = Camera.right[1] * Camera.forward[2] - Camera.right[2] * Camera.forward[1];
				Camera.up[1] = Camera.right[2] * Camera.forward[0] - Camera.right[0] * Camera.forward[2];
				Camera.up[2] = Camera.right[0] * Camera.forward[1] - Camera.right[1] * Camera.forward[0];
			};
			OGLCamera GetCamera()
			{
				return Camera;
			};


			COGLIBAPI VideoReset()
			{
				Win32::Renderer::Get().resetting = 1;
				Win32::Renderer::Get().Shutdown();
				Win32::Renderer::Get().resetting = 0;
				Win32::Renderer::Get().Startup();
				Win32::Renderer::Get().ReloadTextures();
				return new COG::VMType(COG::VMTC_VOID);
			}
		};
	};
};

#endif // _OGL_RENDER_H_INCLUDED_