/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _QUATERNION_H_INCLUDED_
#define _QUATERNION_H_INCLUDED_

#include "vector.h"

namespace Sith
{
	namespace Math
	{
		class Quaternion
		{
		public:
			float x, y, z, w;

			Quaternion()
			{
			}
			// removed normalization of axis and xyzw from set
			Quaternion(const Vector &axis, float radians)
			{
				radians *= .5;
				w = cos(radians);
				float sr = sin(radians);
				x = axis.x * sr;
				y = axis.y * sr;
				z = axis.z * sr;
			}
			void Set(const Vector &axis, float radians)
			{
				radians *= .5;
				w = cos(radians);
				float sr = sin(radians);
				x = axis.x * sr;
				y = axis.y * sr;
				z = axis.z * sr;
			}
			Quaternion(const float *a, float radians)
			{
				radians *= .5;
				w = cos(radians);
				float sr = sin(radians);
				x = a[0] * sr;
				y = a[1] * sr;
				z = a[2] * sr;
			}
			void Set(const float *a, float radians)
			{
				radians *= .5;
				w = cos(radians);
				float sr = sin(radians);
				x = a[0] * sr;
				y = a[1] * sr;
				z = a[2] * sr;
			}

			Quaternion operator * (const Quaternion &param)
			{
				Vector v1, v2, vp;
				Quaternion qp;

				v1.x = x; v1.y = y; v1.z = z;
				v2.x = param.x; v2.y = param.y; v2.z = param.z;
				qp.w = w * param.w - VectorDot(v1, v2);
				vp = v2 * w + v1 * param.w - VectorCross(v1, v2);
				qp.x = vp.x;
				qp.y = vp.y;
				qp.z = vp.z;
				return qp;
			}
			void operator *= (const Quaternion &param)
			{
				float vp[3];

				vp[0] = param.x * w + x * param.w - (y * param.z - z * param.y);
				vp[1] = param.y * w + y * param.w - (z * param.x - x * param.z);
				vp[2] = param.z * w + z * param.w - (x * param.y - y * param.x);

				w = w * param.w - (x * param.x + y * param.y + z * param.z);

				x = vp[0];
				y = vp[1];
				z = vp[2];
			}
			// fills a 4x4 matrix
			void Matrix(float *m)
			{
				float x2, y2, z2, xx, yy, zz, xy, yz, xz, wx, wy, wz;
				x2 = x + x;
				y2 = y + y;
				z2 = z + z;
				xx = x * x2;
				yy = y * y2;
				zz = z * z2;
				xy = x * y2;
				yz = y * z2;
				xz = z * x2;
				wx = w * x2;
				wy = w * y2;
				wz = w * z2;
				m[0] = 1 - (yy + zz);
				m[1] = xy + wz;
				m[2] = xz - wy;
				m[3] = 0;
				m[4] = xy - wz;
				m[5] = 1 - (xx + zz);
				m[6] = yz + wx;
				m[7] = 0;
				m[8] = xz + wy;
				m[9] = yz - wx;
				m[10] = 1 - (xx + yy);
				m[11] = 0;
				m[12] = 0;
				m[13] = 0;
				m[14] = 0;
				m[15] = 1;
			}
			Quaternion Slerp(const Quaternion &q, const float c)
			{
				float omega, cosfomega, sinfomega, k1, k2;
				Quaternion q2, q3;
				cosfomega = x * q.x + y * q.y + z * q.z + w * q.w;
				if (cosfomega < 0)
				{
					cosfomega = -cosfomega;
					q2.x = -q.x;
					q2.y = -q.y;
					q2.z = -q.z;
					q2.w = -q.w;
				}
				else
					q2 = q;

				if (1 - cosfomega > .000001)
				{
					omega = acos(cosfomega);
					sinfomega = sin(omega);
					k1 = sin((1 - c) * omega) / sinfomega;
					k2 = sin(c * omega) / sinfomega;
				}
				else
				{
					k1 = 1 - c;
					k2 = c;
				}

				q3.x = x * k1 + q2.x * k2;
				q3.y = y * k1 + q2.y * k2;
				q3.z = z * k1 + q2.z * k2;
				q3.w = w * k1 + q2.w * k2;

				return q3;
			}

			Quaternion& Conjugate()
			{
				x *= -1;
				y *= -1;
				z *= -1;
				return *this;
			}
		};
	};
};

#endif // _QUATERNION_H_INCLUDED_